import pytest
from src.dirigera.hub.abstract_smart_home_hub import FakeDirigeraHub
from src.dirigera.devices.light import dict_to_light
from src.dirigera.devices.light import Light, StartupEnum


@pytest.fixture(name="fake_client")
def fixture_fake_client():
    return FakeDirigeraHub()


@pytest.fixture(name="fake_light")
def fixture_light(fake_client: FakeDirigeraHub):
    return Light(
        dirigera_client=fake_client,
        light_id="abcd",
        is_reachable=True,
        custom_name="good lamp",
        is_on=True,
        startup_on_off=StartupEnum.START_ON,
        light_level=20,
        color_temp=45,
        color_temp_min=4000,
        color_temp_max=2000,
        color_hue=200,
        color_saturation=0.7,
        can_receive=[
            "customName",
            "isOn",
            "lightLevel",
            "colorTemperature",
            "colorHue",
            "colorSaturation",
        ],
    )


def test_set_name(fake_light: Light, fake_client: FakeDirigeraHub):
    new_name = "stadtlampefluss"
    fake_light.set_name(new_name)
    action = fake_client.patch_actions.pop()
    assert action["route"] == f"/devices/{fake_light.light_id}"
    assert action["data"] == [{"attributes": {"customName": new_name}}]
    assert fake_light.custom_name == new_name


def test_set_light_on(fake_light: Light, fake_client: FakeDirigeraHub) -> None:
    fake_light.set_light(True)
    action = fake_client.patch_actions.pop()
    assert action["route"] == f"/devices/{fake_light.light_id}"
    assert action["data"] == [{"attributes": {"isOn": True}}]
    assert fake_light.is_on


def test_set_light_off(fake_light: Light, fake_client: FakeDirigeraHub) -> None:
    fake_light.set_light(False)
    action = fake_client.patch_actions.pop()
    assert action["route"] == f"/devices/{fake_light.light_id}"
    assert action["data"] == [{"attributes": {"isOn": False}}]
    assert not fake_light.is_on


def test_set_light_level(fake_light: Light, fake_client: FakeDirigeraHub) -> None:
    level = 80
    fake_light.set_light_level(level)
    action = fake_client.patch_actions.pop()
    assert action["route"] == f"/devices/{fake_light.light_id}"
    assert action["data"] == [{"attributes": {"lightLevel": level}}]
    assert fake_light.light_level == level


def test_set_color_temperature(fake_light: Light, fake_client: FakeDirigeraHub) -> None:
    temp = 2200
    fake_light.set_color_temperature(temp)
    action = fake_client.patch_actions.pop()
    assert action["route"] == f"/devices/{fake_light.light_id}"
    assert action["data"] == [{"attributes": {"colorTemperature": temp}}]
    assert fake_light.color_temp == temp


def test_set_light_color(fake_light: Light, fake_client: FakeDirigeraHub) -> None:
    hue = 120
    saturation = 0.9
    fake_light.set_light_color(hue, saturation)
    action = fake_client.patch_actions.pop()
    assert action["route"] == f"/devices/{fake_light.light_id}"
    assert action["data"] == [
        {"attributes": {"colorHue": hue, "colorSaturation": saturation}}
    ]
    assert fake_light.color_hue == hue
    assert fake_light.color_saturation == saturation


def test_set_startup_behaviour_off(
    fake_light: Light, fake_client: FakeDirigeraHub
) -> None:
    behaviour = StartupEnum.START_OFF
    fake_light.set_startup_behaviour(behaviour)
    action = fake_client.patch_actions.pop()
    assert action["route"] == f"/devices/{fake_light.light_id}"
    assert action["data"] == [{"attributes": {"startupOnOff": behaviour}}]
    assert fake_light.startup_on_off == behaviour


def test_set_startup_behaviour_on(
    fake_light: Light, fake_client: FakeDirigeraHub
) -> None:
    behaviour = StartupEnum.START_ON
    fake_light.set_startup_behaviour(behaviour)
    action = fake_client.patch_actions.pop()
    assert action["route"] == f"/devices/{fake_light.light_id}"
    assert action["data"] == [{"attributes": {"startupOnOff": behaviour}}]
    assert fake_light.startup_on_off == behaviour


def test_dict_to_light(fake_client: FakeDirigeraHub):
    data = {
        "id": "1237-343-2dfa",
        "type": "light",
        "deviceType": "light",
        "isReachable": True,
        "attributes": {
            "customName": "Light 2",
            "isOn": False,
            "startupOnOff": "startOn",
            "lightLevel": 100,
            "colorTemperature": 2710,
            "colorTemperatureMin": 4000,
            "colorTemperatureMax": 2202,
            "colorMode": "temperature",
        },
        "capabilities": {
            "canSend": [],
            "canReceive": ["customName", "isOn", "lightLevel"],
        },
    }

    light = dict_to_light(data, fake_client)
    assert light.dirigera_client == fake_client
    assert light.light_id == data["id"]
    assert light.is_reachable == data["isReachable"]
    assert light.custom_name == data["attributes"]["customName"]
    assert light.is_on == data["attributes"]["isOn"]
    assert light.startup_on_off == data["attributes"]["startupOnOff"]
    assert light.light_level == data["attributes"]["lightLevel"]
    assert light.color_temp == data["attributes"]["colorTemperature"]
    assert light.color_temp_min == data["attributes"]["colorTemperatureMin"]
    assert light.color_temp_max == data["attributes"]["colorTemperatureMax"]
    assert light.can_receive == data["capabilities"]["canReceive"]



import ssl
from typing import Any, Dict, List
import requests
import websocket
from urllib3.exceptions import InsecureRequestWarning

from src.dirigera.hub.abstract_smart_home_hub import AbstractSmartHomeHub
from src.dirigera.devices.light import dict_to_light
from src.dirigera.devices.light import Light

requests.packages.urllib3.disable_warnings(  # pylint: disable=no-member
    category=InsecureRequestWarning
)


class Hub(AbstractSmartHomeHub):
    def __init__(
        self,
        token: str,
        ip_address: str,
        port: str = "8443",
        api_version: str = "v1",
    ) -> None:
        self.api_base_url = f"https://{ip_address}:{port}/{api_version}"
        self.websocket_base_url = f"wss://{ip_address}:{port}/{api_version}"
        self.token = token

    def headers(self):
        return {"Authorization": f"Bearer {self.token}"}

    def create_event_listener(
        self,
        on_open: Any | None = None,
        on_message: Any | None = None,
        on_error: Any | None = None,
        on_close: Any | None = None,
        on_ping: Any | None = None,
        on_pong: Any | None = None,
        on_data: Any | None = None,
        on_cont_message: Any | None = None,
    ):
        wsapp = websocket.WebSocketApp(
            self.websocket_base_url,
            header={"Authorization": f"Bearer {self.token}"},
            on_open=on_open,
            on_message=on_message,
            on_error=on_error,
            on_close=on_close,
            on_ping=on_ping,
            on_pong=on_pong,
            on_data=on_data,
            on_cont_message=on_cont_message,
        )

        wsapp.run_forever(sslopt={"cert_reqs": ssl.CERT_NONE})

    def patch(self, route: str, data: Dict[str, Any]) -> Dict[str, Any]:
        response = requests.patch(
            f"{self.api_base_url}{route}",
            headers=self.headers(),
            json=data,
            timeout=10,
            verify=False,
        )
        response.raise_for_status()
        return response.text

    def get(self, route: str) -> Dict[str, Any]:
        response = requests.get(
            f"{self.api_base_url}{route}",
            headers=self.headers(),
            timeout=10,
            verify=False,
        )
        response.raise_for_status()
        return response.json()

    def get_lights(self) -> List[Light]:
        """
        Fetches all lights registered in the Hub
        """
        devices = self.get("/devices")
        lights = list(filter(lambda x: x["type"] == "light", devices))
        return [dict_to_light(light, self) for light in lights]

    def get_light_by_name(self, lamp_name: str) -> Light:
        """
        Fetches all lights and returns first result that matches this name
        """
        lights = self.get_lights()
        lights = list(filter(lambda x: x.custom_name == lamp_name, lights))
        if len(lights) == 0:
            raise AssertionError(f"No light found with name {lamp_name}")
        return lights[0]



from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List

from src.dirigera.hub.abstract_smart_home_hub import AbstractSmartHomeHub


class StartupEnum(Enum):
    START_ON = "startOn"
    START_OFF = "startOff"


@dataclass
class Light:
    dirigera_client: AbstractSmartHomeHub
    light_id: str
    is_reachable: bool
    custom_name: str
    is_on: bool
    startup_on_off: StartupEnum
    light_level: int | None
    color_temp: int | None
    color_temp_min: int | None
    color_temp_max: int | None
    color_hue: int | None
    color_saturation: float | None
    can_receive: List[str]

    def refresh(self) -> None:
        data = self.dirigera_client.get(route=f"/devices/{self.light_id}")
        attributes: Dict[str, Any] = data["attributes"]
        self.light_id=data["id"]
        self.is_reachable=data["isReachable"]
        self.custom_name=attributes["customName"]
        self.is_on=attributes["isOn"]
        self.startup_on_off=attributes["startupOnOff"]
        self.light_level=attributes.get("lightLevel")
        self.color_temp=attributes.get("colorTemperature")
        self.color_temp_min=attributes.get("colorTemperatureMin")
        self.color_temp_max=attributes.get("colorTemperatureMax")
        self.color_hue=attributes.get("colorHue")
        self.color_saturation=attributes.get("colorSaturation")
        self.can_receive=data["capabilities"]["canReceive"]

    def set_name(self, name: str) -> None:
        if "customName" not in self.can_receive:
            raise AssertionError("This lamp does not support the swith-off function")

        data = [{"attributes": {"customName": name}}]
        self.dirigera_client.patch(route=f"/devices/{self.light_id}", data=data)
        self.custom_name = name

    def set_light(self, lamp_on: bool) -> None:
        if "isOn" not in self.can_receive:
            raise AssertionError("This lamp does not support the swith-off function")

        data = [{"attributes": {"isOn": lamp_on}}]
        self.dirigera_client.patch(route=f"/devices/{self.light_id}", data=data)
        self.is_on = lamp_on

    def set_light_level(self, light_level: int) -> None:
        if "lightLevel" not in self.can_receive:
            raise AssertionError(
                "This lamp does not support the set lightLevel function"
            )
        if light_level < 0 or light_level > 100:
            raise AssertionError("light_level must be a value between 0 and 100")

        data = [{"attributes": {"lightLevel": light_level}}]
        self.dirigera_client.patch(route=f"/devices/{self.light_id}", data=data)
        self.light_level = light_level

    def set_color_temperature(self, color_temp: int) -> None:
        if "colorTemperature" not in self.can_receive:
            raise AssertionError(
                "This lamp does not support the set colorTemperature function"
            )
        if color_temp < self.color_temp_max or color_temp > self.color_temp_min:
            raise AssertionError(
                f"color_temperature must be a value between {self.color_temp_max} and {self.color_temp_min}"
            )

        data = [{"attributes": {"colorTemperature": color_temp}}]
        self.dirigera_client.patch(route=f"/devices/{self.light_id}", data=data)
        self.color_temp = color_temp

    def set_light_color(self, hue: int, saturation: float) -> None:
        if (
            "colorHue" not in self.can_receive
            or "colorSaturation" not in self.can_receive
        ):
            raise AssertionError(
                "This lamp does not support the set light color function"
            )
        if hue < 0 or hue > 360:
            raise AssertionError("hue must be a value between 0 and 360")
        if saturation < 0.0 or saturation > 1.0:
            raise AssertionError("saturation must be a value between 0.0 and 1.0")

        data = [{"attributes": {"colorHue": hue, "colorSaturation": saturation}}]
        self.dirigera_client.patch(route=f"/devices/{self.light_id}", data=data)
        self.color_hue = hue
        self.color_saturation = saturation

    def set_startup_behaviour(self, behaviour: StartupEnum) -> None:
        """
        Sets the behaviour of the lamp in case of a power outage.
        When set to START_ON the lamp will turn on once the power is back.
        When set to START_OFF the lamp will stay off once the power is back.
        """
        data = [{"attributes": {"startupOnOff": behaviour}}]
        self.dirigera_client.patch(route=f"/devices/{self.light_id}", data=data)
        self.startup_on_off = behaviour


def dict_to_light(data: Dict[str, Any], dirigera_client: AbstractSmartHomeHub):
    attributes: Dict[str, Any] = data["attributes"]
    light_level = attributes.get("lightLevel")
    color_temp = attributes.get("colorTemperature")
    color_temp_min = attributes.get("colorTemperatureMin")
    color_temp_max = attributes.get("colorTemperatureMax")
    color_hue = attributes.get("colorHue")
    color_saturation = attributes.get("colorSaturation")

    return Light(
        dirigera_client=dirigera_client,
        light_id=data["id"],
        is_reachable=data["isReachable"],
        custom_name=attributes["customName"],
        is_on=attributes["isOn"],
        startup_on_off=attributes["startupOnOff"],
        light_level=light_level,
        color_temp=color_temp,
        color_temp_min=color_temp_min,
        color_temp_max=color_temp_max,
        color_hue=color_hue,
        color_saturation=color_saturation,
        can_receive=data["capabilities"]["canReceive"],
    )



import abc
from typing import Any, Dict, List


class AbstractSmartHomeHub(abc.ABC):
    @abc.abstractmethod
    def patch(self, route: str, data: Dict[str, Any]) -> Dict[str, Any]:
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, route: str) -> Dict[str, Any]:
        raise NotImplementedError


class FakeDirigeraHub(AbstractSmartHomeHub):
    def __init__(self) -> None:
        self.patch_actions: List = []
        self.get_actions: List = []

    def patch(self, route: str, data: Dict[str, Any]) -> Dict[str, Any]:
        self.patch_actions.append({"route": route, "data": data})

    def get(self, route: str) -> Dict[str, Any]:
        self.get_actions.append({"route": route})
